/*
 *
 * Buffering of Messages
 *
 * Author:
 * Ricardo Baratto
 * Network Computing Lab - Columbia University
 *
 */

/* set tabstop = 8 */
#ifndef _THINC_BUFFER_H_
#define _THINC_BUFFER_H_

/* How it works:
 * When a commands need to be sent to all clients, the command is
 * appropiately clipped to each client's clip, but instead of sending
 * out the wire it is stored in the client's out buffer. Later,
 * (possibly in the BlockHandler()) the command is flushed out along with
 * all the other commands stored in the queue. However, if flushing is
 * not possible because it would make the server block, the command
 * is left in the queue until a new attempt can be made. During this
 * time the command may need to be modified (partially overwritten) or
 * purged.
 *
 * We use the following command functions:
 * Create() => for commands coming from onscreen drawing operations
 * Destroy() => for commands which are succesfully flushed or purged
 * Modify() => on initial buffering if a command needs to be clipped, and 
 * for commands which are partially overwritten by new commands
 * Flush() => to write commands out the wire
 * Execute() => used to transfer from offscreen pixmaps to the buffer
 * Copy() => Need a copy of each command per client queue
 *
 * The flow of control looks something like this:
 *
 * [ DRAW ] -------> onscreen? --- no ---> [ QueueCmd() ] -----|
 *                       |                                     | Execute()
 *                       |---- yes ---> [ BufferCmd() ] <-------
 *                                            |          
 *                                            |
 *                                            v
 *                                       [ FlushCmd() ]
 *                                        
 * Execute() is triggered by a CopyArea to an onscreen drawable.
 * BufferCmd() has two cases:
 * => if command comes from offscreen, not much needs to be done, just
 * clipping to the client's region
 * => if command comes directly from draw operation, a new command struct
 * needs to be created (Create() func pointer)
 *
 * If buffering is disabled, BufferCmd() just calls the Flush() function
 * directly, making everything transparent to the command-specific 
 * functions.
 */
int thincBufferCmd(thincScreenPtr pThinc, WindowPtr pWindow, thincCmdPtr cmd);
int thincBufferCmdNoCopy(thincScreenPtr pThinc, WindowPtr pWindow, 
	thincCmdPtr cmd);
int thincBufferNewCmd(thincScreenPtr pThinc, WindowPtr pWindow, int type, 
	int x, int y, int w, int h, ...);
int thincBufferNewCmdReg(thincScreenPtr pThinc, WindowPtr pWindow, int type, 
	RegionPtr reg, ...);
int thincFlushCmdAll(thincScreenPtr pThinc);
int thincFlushCmdClient(thincScreenPtr pThinc, thincClientPtr cl);
int thincFlushListBuffer(thincScreenPtr pThinc, thincClientPtr cl, 
	thincListPtr buffer);
int thincFlushListSched(thincScreenPtr pThinc, thincClientPtr cl, 
	thincListPtr queue);

int thincBufferCmdClientReal(thincClientPtr cl, thincCmdPtr cmd);
int thincFlushCmds(thincScreenPtr pThinc, thincClientPtr cl);

int thincBufferBatchStart(thincScreenPtr pThinc, RegionPtr reg);
int thincBufferBatchEnd(thincScreenPtr pThinc);

int thincSendPingAll(thincScreenPtr pThinc);

#ifdef T_NODRAW
int thincClient0Buffer(thincClientPtr cl, thincCmdPtr cmd);
int thincClient0Flush(thincScreenPtr pThinc, thincClientPtr cl);
int thincClient0SendFB(thincScreenPtr pThinc, thincClientPtr cl);
#endif

#endif	/* _THINC_BUFFER_H_ */
