/* Copyright 2003, 2004, 2005 Ricardo Baratto */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
 *
 * Functions for offscreen manipulation of RAW commands
 *
 * Author:
 * Ricardo Baratto
 * Network Computing Lab - Columbia University
 *
 */


/* set tabstop = 8 */

#include "thinc.h"

/* for write to file function */
#if 0
#define T_ENABLE_WRITEFILE_FUNC
#endif
/* to use it in this code */
#if 0
#ifndef T_WIN32
#define T_ENABLE_WRITEFILE
#endif
#endif

#ifdef T_ENABLE_WRITEFILE_FUNC
#ifndef T_WIN32
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <netinet/in.h>
#endif
#endif

static int thincRawMerge(thincScreenPtr pThinc, thincCmdPtr cmd,
	thincCmdPtr merge);
static thincCmdPtr thincRawSave(thincScreenPtr pThinc, thincCmdPtr cmd);
static int thincRawClipData(thincScreenPtr pThinc, struct rawCmd *raw);
static int thincRawMergeData(thincScreenPtr pThinc, struct rawCmd *raw,
	struct rawCmd *mraw, BoxPtr merged, int where);

#define T_RAW_CMDSIZE	sizeof(struct rawCmd)

/* Raw Cmd is stored as:
 * -------------------
 * | CmdRec | struct |
 * |        | rawCmd |
 * -------------------
 *               |--------> RegionData (may be NULL)
 *
 * SAVED PIXEL DATA:
 * In addition, a command may have pixel data saved with it. This is stored
 * outside of the main structure. Since trying to maintain this data in sync
 * with the rectangles in the Region would be a nightmare, we do the 
 * following: We will store the data for the extents of the original 
 * region, not just for each rectangle in the region. When the data is
 * needed, we just need to figure out where the data for each rectangle
 * resides (offset from orig) and take it out. The process goes along the
 * lines of:
 * - On create, we save the original extents in rawCmd.orig
 * - If we're clipped and no data has been saved, the original extents are
 *   modified appropriately.
 * - If we're clipped and data has been saved, the original extents _and_ the
 *   data are untouched.
 * - If we're translated, the extents are modified appropriately (note that
 *   we work in terms of offsets so this doesn't affect anything).
 * - If we're merged then the original extents are extended appropriately.
 *   Since currently we can only merge 1-rectangle commands, then merging
 *   the data is (almost) trivial.
 * - At flush time, offsets for each rectangle are computed and the
 *   data extracted.
 */
thincCmdPtr thincRawCreate(thincScreenPtr pThinc, RegionPtr pReg, 
	va_list data)
{
	t_dbg_name(func)
	int cmdsize, totalsize;
	unsigned int datasize;

	/* the command */
	thincCmdPtr cmd;
	struct rawCmd *raw;
	char *rawcmd;
	stats_var2(t1, t2);
	stats_start(pThinc, t1);

	thincDbg(T_DBG_CMD, ("%s: size of rawCmd: %d bytes\n", func,
		sizeof(struct rawCmd)));

	cmdsize = T_RAW_CMDSIZE;

	totalsize = cmdsize + T_CMDREC_SIZE;

	thincDbg(T_DBG_CMD, ("%s: command size %d, total size %d bytes\n", 
		func, cmdsize, totalsize));
	rawcmd  = malloc(totalsize);
	if(!rawcmd) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't allocate "
			"memory for RAW command\n", func));
		return NULL;
	}

	/* init the command structure */
	cmd = (thincCmdPtr) rawcmd;
	if(thincRawInit(pThinc, cmd, pReg) == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't initialize "
			"RAW command\n", func));
		free(rawcmd);
		return NULL;
	}

	/* fill rawCmd */
	raw = cmd_data(cmd, struct rawCmd*);
	datasize = va_arg(data, unsigned int);

	if(datasize) {
		thincDbg(T_DBG_CMD, ("%s: creating RAW command with data "
			"size %u\n", func, datasize));
		raw->datasize = datasize;
		/* it needs to be a pointer so we can share it among
		 * different commands */
		raw->data = malloc(sizeof(rawData));
		if(!raw->data) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't allocate "
				"memory for RAW command data\n", func));
			REGION_UNINIT(pThinc->pScreen, &raw->reg);
			free(rawcmd);
			return NULL;
		}
		raw->data->refcount = 1;
		raw->data->data = va_arg(data, unsigned char*);
	}

	/* the end */
	thincDbg(T_DBG_CMD, ("%s: finished, command's id %#lx\n", func,
		cmd->id));

	stats_end(pThinc, t1, t2, rawStats.create);

	return cmd;
}

int thincRawInit(thincScreenPtr pThinc, thincCmdPtr cmd, RegionPtr pReg)
{
	struct rawCmd *raw;
	BoxPtr ext;

	cmd->type = T_FB_RUPDATE;
	cmd->flags = 0;
	cmd->id = thinc_getCmdID();
	ext = REGION_EXTENTS(pThinc->pScreen, pReg);
	copy_boxp(&cmd->area, ext);
	cmd->cmdSize = T_RAW_CMDSIZE;

	raw = cmd_data(cmd, struct rawCmd*);

	copy_boxp(&raw->orig, ext);

	raw->datasize = 0;
	raw->data = NULL;

	raw->reg = thincNullReg; /* so that REGION_COPY doesn't crash */ 
	if(!REGION_COPY(pThinc->pScreen, &raw->reg, pReg))
		return T_ERROR;

#ifdef T_USERENDER
	raw->Bpp = pThinc->Bpp;
#endif

	return T_OK;
}

/* used when punting from resized glyphs */
thincCmdPtr thincRawCreateV(thincScreenPtr pThinc, RegionPtr pReg, ...)
{
	va_list args;
	thincCmdPtr cmd;

	va_start(args, pReg);
	cmd = thincRawCreate(pThinc, pReg, args);
	va_end(args);

	return cmd;
}

void thincRawDestroy(thincScreenPtr pThinc, thincCmdPtr cmd)
{
	t_dbg_name(func)
	char *rawcmd = (char*)cmd;
	struct rawCmd *raw;
	stats_var2(t1, t2);

	stats_start(pThinc, t1);

	thincDbg(T_DBG_CMD, ("%s: destroying command %#lx\n", func, 
		cmd->id));

	raw = cmd_data(cmd, struct rawCmd*);
	REGION_UNINIT(pThinc->pScreen, &raw->reg);

	if(raw->datasize) {
		t_bug_on_2(
			(!raw->data->refcount),
			T_BUG_ARG("Found RAW command %#lx with 0 reference "
				"count!!\n", cmd->id),
			/* if we free it, we risk segfaulting.. what to
			 * do? */
			return
		);
		raw->data->refcount--;
		if(!raw->data->refcount) {
			thincDbg(T_DBG_CMD, ("%s: destroying RAW command "
				"with data and 0 reference count\n", func));
			free(raw->data->data);
			free(raw->data);
			raw->data = NULL;
		} else {
			thincDbg(T_DBG_CMD, ("%s: command with non-zero "
				"refcount %u, not freeing its data\n", func,
				raw->data->refcount));
		}
	}

	free(rawcmd);

	thincDbg(T_DBG_CMD, ("%s: finished\n", func));
	stats_end(pThinc, t1, t2, rawStats.destroy);
}

thincCmdPtr thincRawCopy(thincScreenPtr pThinc, thincCmdPtr cmd)
{
	t_dbg_name(func)

	char *cnewcmd, *coldcmd;
	int totalsize;

	struct rawCmd *nraw, *oraw;
	thincCmdPtr ncmd;
	stats_var2(t1, t2);

	stats_start(pThinc, t1);

	thincDbg(T_DBG_CMD, ("%s: Copying command %#lx of size %d bytes\n",
		func, cmd->id, cmd->cmdSize));

	totalsize = cmd->cmdSize + T_CMDREC_SIZE;

	cnewcmd = malloc(totalsize);
	if(!cnewcmd) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't allocate memory "
			"for copy of RAW command\n", func));
		return NULL;
	}

	/* easier to just memcpy */
	coldcmd = (char*)cmd;
	memcpy(cnewcmd, coldcmd, totalsize);

	ncmd = (thincCmdPtr) cnewcmd;
	nraw = cmd_data(ncmd, struct rawCmd*);
	oraw = cmd_data(cmd, struct rawCmd*); 

	if(oraw->datasize) {
		nraw->datasize = oraw->datasize;
		oraw->data->refcount++;
		nraw->data = oraw->data;
		thincDbg(T_DBG_CMD, ("%s: got RAW command with data. "
			"its refcount is now %u\n", func, 
			nraw->data->refcount));
#ifdef T_ENABLE_WRITEFILE
		thincRawWriteToFile(oraw->data->data, box_width(&oraw->orig),
			box_height(&oraw->orig), pThinc->Bpp, 
			box_width_bytes(pThinc, &oraw->orig));
		thincRawWriteToFile(nraw->data->data, box_width(&nraw->orig),
			box_height(&nraw->orig), pThinc->Bpp, 
			box_width_bytes(pThinc, &nraw->orig));
#endif
	} else {
		nraw->datasize = 0;
		nraw->data = NULL;
	}

	nraw->reg = thincNullReg;
	if(!REGION_COPY(pThinc->pScreen, &nraw->reg, &oraw->reg)) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't copy region "
			"of RAW command\n", func));
		if(nraw->data) {
			oraw->data->refcount--;
			nraw->data = NULL;
		}
		free(cnewcmd);
		return NULL;
	}

	thincDbg(T_DBG_CMD, ("%s: finished\n", func));

	stats_end(pThinc, t1, t2, rawStats.copy);
	return ncmd;
}

/* If mod is clip & translate, clipping is done first.
 * If after modification region is empty, return NULL. Otherwise, return
 * modified command */
thincCmdPtr thincRawModify(thincScreenPtr pThinc, thincCmdPtr cmd, 
	int mod, va_list args)
{
	t_dbg_name(func)

	struct rawCmd *raw;
	RegionPtr reg;

	int deltax, deltay;
	RegionPtr clip;
	thincCmdPtr merge;
	int *merged, havedata = T_FALSE, ret;
	stats_var2(t1, t2);

	stats_start(pThinc, t1);

	thincDbg(T_DBG_CMD, ("%s: modifying cmd %#lx. operations: %d\n",
		func, cmd->id, mod));

	raw = cmd_data(cmd, struct rawCmd*);
	reg = &raw->reg;

	/* TODO */
#if 1
	if(raw->datasize) {
		havedata = T_TRUE;
		thincWarn("Modifying command with saved data. Unsupported. "
			"Modification: %d\n", mod);
	}
#endif

	/* clip? */
	if(mod & (T_CMD_CLIP | T_CMD_OVER)) {
		clip = va_arg(args, RegionPtr);
		REGION_INTERSECT(pThinc->pScreen, reg, clip, reg);
		if(!(REGION_NOTEMPTY(pThinc->pScreen, reg))) {
			thincDbg(T_DBG_CMD, 
				("%s: command has been completely "
				"obscured. finished\n", func));
			return NULL;
		}
		if(!havedata) {
			copy_boxp(&raw->orig, 
				(REGION_EXTENTS(pThinc->pScreen, reg)));
		}

		thincDbg(T_DBG_CMD, ("%s: clipped the command, it now "
			"has %ld rectangles\n", func, REGION_NUM_RECTS(reg)));
		print_reg(T_DBG_CMD,pThinc->pScreen,reg,"clipped region:");
	}

	/* translate? */
	if(mod & T_CMD_MOVE) {
		deltax = va_arg(args, int);
		deltay = va_arg(args, int);
		box_translate(&raw->orig, deltax, deltay);
		REGION_TRANSLATE(pThinc->pScreen, reg, deltax, deltay);
		
		print_reg(T_DBG_CMD, pThinc->pScreen, reg,
			"translated region:");
	}

	if(mod & T_CMD_MERGE) {
		merge = va_arg(args, thincCmdPtr);
		merged = va_arg(args, int*);
		if( (ret = thincRawMerge(pThinc, cmd, merge)) == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't merged "
				"commands\n", func));
			return (thincCmdPtr) T_ERROR;
		} else if(ret == T_TRUE) {
			*merged = T_TRUE;
			print_reg(T_DBG_CMD, pThinc->pScreen, reg, 
				"merged region: ");
		} else { /* ret == T_FALSE */
			thincDbg(T_DBG_CMD, ("%s: Couldn't merge "
				"commands. Sorry\n", func));
			*merged = T_FALSE;
		}
	}

	if(mod & T_CMD_SAVE) {
		cmd = thincRawSave(pThinc, cmd);
		if(cmd == (thincCmdPtr) T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't save "
				"command's data\n", func));
			return (thincCmdPtr) T_ERROR;
		}
	}

	/* update the bounding box */
	copy_boxp(&cmd->area, (REGION_EXTENTS(pThinc->pScreen, reg)));

	/* finished */
	thincDbg(T_DBG_CMD, ("%s: finished\n", func));
	stats_end(pThinc, t1, t2, rawStats.modify);

	return cmd;
}

/**** try to merge command merge with cmd ****/

/* Returns:
 * T_ERROR on error
 * T_TRUE if commands were merged
 * T_FALSE if commands couldn't be merged
 */
static int thincRawMerge(thincScreenPtr pThinc, thincCmdPtr cmd,
	thincCmdPtr merge) 
{
	t_dbg_name(func)
	struct rawCmd *raw, *mraw;
	RegionPtr reg, mreg;
	BoxPtr rect, mrect;
	int ret;
	int err, havedata, 
		where = T_MERGE_TOP; /* stupid compiler warnings */
	BoxRec copy;
	BoxPtr real;

	thincDbg(T_DBG_CMD, ("%s: trying to merge command %#lx into command "
		"%#lx\n", func, merge->id, cmd->id));

	raw = cmd_data(cmd, struct rawCmd*);
	mraw = cmd_data(merge, struct rawCmd*);
	reg = &raw->reg;
	mreg = &mraw->reg;

	/* both must have data */
	if( ( (!raw->datasize) && (mraw->datasize) ) ||
	    ( (raw->datasize) && (!mraw->datasize) ) ) {
		thincDbg(T_DBG_CMD, ("%s: one command has data but the "
			"other doesn't. don't know how to merge those, "
			"sorry. finished\n", func));
		return T_FALSE;
	}
	havedata = T_TRUE;	/* common case */
	if(!raw->datasize)
		havedata = T_FALSE;

	if( (REGION_NUM_RECTS(reg)) != 1 || (REGION_NUM_RECTS(mreg) != 1)) {
		/* FIXME: We should try to merge these too */
		thincDbg(T_DBG_CMD, ("%s: commands are not "
			"simple enough. Sorry, can't merge\n", func));
		return T_FALSE;
	}
	rect = REGION_RECTS(reg);
	/* need to make rect point to a copy of the rectangle in the region.
	 * we need the original later on */
	real = rect;
	copy_boxp(&copy, rect);
	rect = &copy;
	mrect = REGION_RECTS(mreg);

	print_box(T_DBG_CMD, mrect, "command to merge: ");
	print_box(T_DBG_CMD, rect, "command to merge with: ");

	ret = thincMergeBoxes(rect, mrect, &where);
	if(!ret)
		goto end;

	/* if we merged and commands have data, then:
	 * -> clip both command's data.
	 * -> merge appropriately (using where)
	 * this is very expensive. we should avoid this at all
	 * costs */
	if(havedata) {
		thincDbg(T_DBG_CMD, ("%s: merging commands' data\n", 
			func));
		err = thincRawClipData(pThinc, raw);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't clip "
				"merged commands' data\n", func));
			/* XXX: this is bad, the commands may be all screwed
			 * up :( */
			return T_ERROR;
		}
		err = thincRawClipData(pThinc, mraw);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't clip "
				"merged commands' data\n", func));
			return T_ERROR;
		}
		err = thincRawMergeData(pThinc, raw, mraw, rect, where);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't merge "
				"commands' data\n", func));
			return T_ERROR;
		}
	}

	thincDbg(T_DBG_CMD, ("%s: got a merge, revalidating region "
		"extents\n", func));

	/* copy merged rect back to region */
	copy_boxp(real, rect);
	rect = real;
	copy_boxp(&reg->extents, rect);
	/* fix orig */
	copy_boxp(&raw->orig, rect);

	print_box(T_DBG_CMD, REGION_EXTENTS(pThinc->pScreen, &raw->reg), 
		"merged REGION extents");

end:
	thincDbg(T_DBG_CMD, ("%s: finished\n", func));
	return ret;
}

/* Sync the data with the current draw region (data is normally sync'ed
 * with raw->orig). It only supports cases where the command represents 1 
 * rectangle. And note that since we're modifying the data, we need to create 
 * a copy of it if its refcount is more than 1 */
static int thincRawClipData(thincScreenPtr pThinc, struct rawCmd *raw)
{
	t_dbg_name(func)
	BoxPtr orig, rect;
	unsigned char *from, *to;
	unsigned int datasize, width, height, pitch, rect_width, i;
	RegionPtr reg;
	rawData *copy = NULL;

	thincDbg(T_DBG_CMD, ("%s: clipping pixel data for RAW command\n",
		func));
	if(!raw->datasize) {
		thincDbg(T_DBG_CMD, ("%s: command doesn't have data, "
			"finished\n", func));
		return T_OK;
	}
	reg = &raw->reg;
	if(REGION_NUM_RECTS(reg) != 1) {
		thincDbg(T_DBG_CMD, ("%s: command covers more than 1 "
			"rectangle (%ld), cannot clip its data\n", func,
			REGION_NUM_RECTS(reg)));
		return T_OK;
	}
	rect = REGION_RECTS(reg);
	orig = &raw->orig;

	if(box_equal(rect, orig) ) {
		/* nothing to clip */
		thincDbg(T_DBG_CMD, ("%s: draw rectangle and original "
			"box are the same, nothing to clip, finished\n",
			func));
		return T_OK;
	}

	print_box(T_DBG_CMD, orig, "original: ");
	print_box(T_DBG_CMD, rect, "current: ");

	/* ok, gotta clip:
	 * -> if refcount is > 1, need to clip to a copy of the data
	 * -> find the start of the draw rectangle.
	 * -> memmove() from there to the start of the data
	 * -> realloc() (if refcount == 1)
	 */
	width = box_width(rect);
	height = box_height(rect);
	datasize = box_size(pThinc, rect);

	thincDbg(T_DBG_CMD, ("%s: clipping from %ux%u to %ux%u. new "
		"datasize %u\n", func, box_width(orig), box_height(orig),
		width, height, datasize));

	if(raw->data->refcount != 1) {
		thincDbg(T_DBG_CMD, ("%s: Data is shared (refcount %u) "
			"creating copy\n", func, raw->data->refcount));
		copy = malloc(sizeof(*copy));
		if(!copy) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't allocate "
				"memory for data copy\n", func));
			return T_ERROR;
		}
		copy->data = malloc(datasize);
		if(!copy->data) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't allocate "
				"memory for data copy\n", func));
			return T_ERROR;
		}
		to = copy->data;
	} else {
		thincDbg(T_DBG_CMD, ("%s: RAW data not shared, reusing it\n",
			func));
		to = raw->data->data;
	}
	pitch = box_width_bytes(pThinc, orig);
	t_get_datapointer(raw->data->data, orig, pitch, pThinc->Bpp, rect,
		from);
	rect_width = box_width_bytes(pThinc, rect);

	thincDbg(T_DBG_CMD, ("%s: got pointers, rect_width is %u, "
		"pitch is %u\n", func, rect_width, pitch));

	for(i=0; i<height; ++i) {
		memmove(to, from, rect_width);
		to += rect_width;
		from += pitch;
	}
	if(raw->data->refcount == 1) {
		/* realloc */
		thincDbg(T_DBG_CMD, ("%s: got all data, realloc'ing to "
			"%u bytes\n", func, datasize));
		raw->data->data = realloc(raw->data->data, datasize);
		if(!raw->data->data) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't realloc "
				"clipped data\n", func));
			return T_ERROR;
		}
	} else {
		thincDbg(T_DBG_CMD, ("%s: got all data, moving copy to "
			"RAW command\n", func));
		/* release it from us, and put copy as our data */
		raw->data->refcount--;
		copy->refcount = 1;
		raw->data = copy;
	}
	raw->datasize = datasize;
	copy_boxp(orig, rect);

	thincDbg(T_DBG_CMD, ("%s: finished\n", func));
	return T_OK;
}

/* copies mraw data to raw's data. Assumes raw and mraw's orig haven't
 * been modified to reflect the merge */
#define copy_alternate(to, from, bytes_line, height, to_pitch) { \
	unsigned int _idx; \
	unsigned char *_fc = (from), *_tc = (to); \
	for(_idx=0; _idx<(height); ++_idx) { \
		memmove(_tc, _fc, (bytes_line)); \
		_tc += (to_pitch); \
		_fc += (bytes_line); \
	} \
}

#define copy_alternate_bottom(to, from, bytes_line, height, to_pitch) { \
	unsigned int _idx; \
	unsigned char *_fc = (from), *_tc = (to); \
	for(_idx=0; _idx<(height); ++_idx) { \
		memmove(_tc, _fc, (bytes_line)); \
		_tc -= (to_pitch); \
		_fc -= (bytes_line); \
	} \
}


static int thincRawMergeData(thincScreenPtr pThinc, struct rawCmd *raw,
	struct rawCmd *mraw, BoxPtr merged, int where)
{
	t_dbg_name(func)
	rawData *rdata, *mrdata, *tordata;
	unsigned char *data, *mdata, *todata;
	unsigned int datasize, freeto;
	unsigned int len, pitch, height;

	thincDbg(T_DBG_CMD, ("%s: merging data for RAW commands\n", func));

	datasize = raw->datasize + mraw->datasize;

	thincDbg(T_DBG_CMD, ("%s: need %u bytes\n", func, datasize));

	rdata = raw->data;
	mrdata = mraw->data;

#ifdef T_ENABLE_WRITEFILE
	thincError("raw data\n");
	thincRawWriteToFile(rdata->data, box_width(&raw->orig), 
		box_height(&raw->orig), pThinc->Bpp, 
		box_width_bytes(pThinc,&raw->orig));
	thincError("mraw data\n");
	thincRawWriteToFile(mrdata->data, box_width(&mraw->orig), 
		box_height(&mraw->orig), pThinc->Bpp, 
		box_width_bytes(pThinc,&mraw->orig));
#endif

	if(rdata->refcount != 1) {
		/* need new one */
		thincDbg(T_DBG_CMD, ("%s: raw data shared (%d), getting "
			"new one\n", func, rdata->refcount));
		tordata = malloc(sizeof(*tordata));
		if(!tordata) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't allocate "
				"memory for merged data\n", func));
			return T_ERROR;
		}
		tordata->refcount = 1;
		tordata->data = NULL;
		rdata->refcount--;
		freeto = T_TRUE;
	} else {
		tordata = raw->data;
		freeto = T_FALSE;
	}

	tordata->data = realloc(tordata->data, datasize);
	if(!tordata->data) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't allocate memory "
			"for merged data\n", func));
		if(freeto) {
			free(tordata);
			rdata->refcount++;
		}
		return T_ERROR;
	}

	thincDbg(T_DBG_CMD, ("%s: realloc'ed data, merging: %d\n", func,
		where));

	todata = tordata->data;
	data = rdata->data;
	mdata = mrdata->data;

	/* IMPORTANT: always copy rdata first or it may get overwritten!! */
	switch(where) {
		case T_MERGE_TOP:
			/* mdata on top, then rdata*/
			memmove(todata + mraw->datasize, data, raw->datasize);
			memmove(todata, mdata, mraw->datasize);
			break;
		case T_MERGE_RIGHT:
			/* data first, then mdata, but alternating lines
			 * and from the bottom to avoid overwriting */
			pitch = box_width_bytes(pThinc, merged);
			len = box_width_bytes(pThinc, &raw->orig);
			height = box_height(&raw->orig);
			
			/* go to bottom line in both */
			todata = tordata->data + datasize - pitch;
			data = rdata->data + raw->datasize - len;
			copy_alternate_bottom(todata, data, len, height, pitch);

			todata = tordata->data + len;
			len = box_width_bytes(pThinc, &mraw->orig);
			copy_alternate(todata, mdata, len, height, pitch);
			break;
		case T_MERGE_BOTTOM:
			/* rdata first, then mdata */
			if(todata != data)
				memmove(todata, data, raw->datasize);
			todata += raw->datasize;
			memmove(todata, mdata, mraw->datasize);
			break;
		case T_MERGE_LEFT:
			/* mdata first, then data, but alternating lines and
			 * data from the bottom to avoid overwritting */
			pitch = box_width_bytes(pThinc, merged);
			len = box_width_bytes(pThinc, &raw->orig);
			height = box_height(&mraw->orig);
			todata = tordata->data + datasize - len;
			data = rdata->data + raw->datasize - len;
			copy_alternate_bottom(todata, data, len, height, pitch);

			todata = tordata->data;
			len = box_width_bytes(pThinc, &mraw->orig);
			copy_alternate(todata, mdata, len, height, pitch);
			break;
		default:
			T_BUG_ARG("Unknown merge position %d passed", where);
			if(freeto) {
				free(tordata);
				rdata->refcount++;
			}
			return T_ERROR;
	}

	/* fix things up */
	raw->datasize = datasize;
	raw->data = tordata;

#ifdef T_ENABLE_WRITEFILE
	thincError("writing merged update\n");
	thincRawWriteToFile(raw->data->data, box_width(merged), 
		box_height(merged), pThinc->Bpp, 
		box_width_bytes(pThinc,merged));
#endif

	thincDbg(T_DBG_CMD, ("%s: finished\n", func));
	return T_OK;
}

static thincCmdPtr thincRawSave(thincScreenPtr pThinc, thincCmdPtr cmd)
{
	t_dbg_name(func)
	struct rawCmd *raw = cmd_data(cmd, struct rawCmd*);
	BoxPtr box = &raw->orig;
	int size, rect_width, fb_width, height, i;
	unsigned char *from, *to;

	thincDbg(T_DBG_CMD, ("%s: saving RAW data for command %#lx\n",
		func, cmd->id));
	print_box(T_DBG_CMD, box, "original box");

	if(raw->datasize) {
		thincDbg(T_DBG_CMD, ("%s: pixel data already saved. "
			"finished\n", func));
		return cmd;
	}

	size = box_size(pThinc, box);
	raw->data = malloc(sizeof(rawData));
	if(!raw->data) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Could not allocate "
			"memory for pixel data\n", func));
		return (thincCmdPtr)T_ERROR;
	}
	raw->data->data = malloc(size);
	if(!raw->data->data) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Could not allocate "
			"memory for pixel data\n", func));
		free(raw->data);
		return (thincCmdPtr)T_ERROR;
	}
	raw->data->refcount = 1;
	raw->datasize = size;

	thincDbg(T_DBG_CMD, ("%s: got memory, copying pixel data\n",
		func));
	to = raw->data->data;
	t_get_fbpointer(pThinc, box, from, rect_width, fb_width);
	height = box_height(box);

	for(i=0; i<height; ++i) {
		memcpy(to, from, rect_width);
		to += rect_width;
		from += fb_width;
	}

	thincDbg(T_DBG_CMD, ("%s: finished\n", func));
	return cmd;
}

void thincRawGetInfo(thincScreenPtr pThinc, thincCmdPtr cmd, int info, ...)
{
	va_list args;
        struct rawCmd *raw = cmd_data(cmd, struct rawCmd*);
	stats_var2(t1, t2);

	stats_start(pThinc, t1);

	va_start(args, info);

	if(info & T_CMD_REGION) {
		RegionRec **reg = va_arg(args, RegionRec**);
		*reg = &raw->reg;
	}

	if(info & T_CMD_SIZE) {
		/* size of this command is:
		 * number of rects * 
		 * 	(header size + T_FB_RUPDATE_SIZE + size of data)
		 */
		unsigned int *size = va_arg(args, unsigned int*);
		unsigned int total = 0, i;
		int numrects = REGION_NUM_RECTS(&raw->reg);
		BoxPtr rects = REGION_RECTS(&raw->reg);
		for(i=0; i<numrects; ++i, ++rects) {
			total += T_FB_RUPDATE_SIZE + THINC_HEADERSIZE;
			total += box_size(pThinc, rects);
		}
		*size = total;
	}

	va_end(args);
	stats_end(pThinc, t1, t2, rawStats.info);
}

void thincRawExecute(thincScreenPtr pThinc, WindowPtr pWindow,
	thincCmdPtr cmd, RegionPtr clip, int deltax, int deltay)
{
	t_dbg_name(func)
	
	struct rawCmd *raw;
	RegionRec bak = thincNullReg;	/* backup of original reg */
	RegionPtr reg;
	stats_var2(t1, t2);
	stats_start(pThinc, t1);

	thincDbg(T_DBG_CMD, ("%s: Executing command %#lx. deltax %d, "
		"deltay %d\n", func, cmd->id, deltax, deltay));
	print_reg(T_DBG_CMD, pThinc->pScreen, clip, "clipping with:");
	thincDbg(T_DBG_CMD, ("%s: clip has %ld rectangles\n", func,
		REGION_NUM_RECTS(clip)));

	raw = cmd_data(cmd, struct rawCmd*);
#ifdef T_ENABLE_WRITEFILE
	thincError("writing command %#lx to file, datasize %u\n", cmd->id,
		raw->datasize);
	thincRawWriteToFile(raw->data, box_width(&raw->orig), 
		box_height(&raw->orig), pThinc->Bpp, 
		box_width_bytes(pThinc, &raw->orig));
#endif

	/* This copy is needed just because we need to do modifications to
	 * the region which are specific to executing the command and
	 * shouldn't be propagated to the original instance of the command.
	 * When the command is buffered, 
	 * TODO: Maybe we can check beforehand if the region will be actually
	 * changed before making the copy */
	if(!REGION_COPY(pThinc->pScreen, &bak, &raw->reg)) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't create copy of "
			"RAW region\n", func));
		return;
	}

	reg = &raw->reg;
	print_reg(T_DBG_CMD, pThinc->pScreen, reg, "raw cmd region:");
	thincDbg(T_DBG_CMD, ("%s: before clip have %ld rectangles\n", func,
		REGION_NUM_RECTS(reg)));

	box_translate(&raw->orig, deltax, deltay);
	REGION_TRANSLATE(pThinc->pScreen, reg, deltax, deltay);
	REGION_INTERSECT(pThinc->pScreen, reg, clip, reg);
	if(!REGION_NOTEMPTY(pThinc->pScreen, reg)) {
		thincDbg(T_DBG_CMD, ("%s: finished. Queued command is "
			"completely obscured\n", func));
		REGION_UNINIT(pThinc->pScreen, reg);
		copy_reg(raw->reg, bak);
		return;
	}

	print_reg(T_DBG_CMD, pThinc->pScreen, reg, "raw cmd after clip:");
	thincDbg(T_DBG_CMD, ("%s: after clip have %ld rectangles\n", func,
		REGION_NUM_RECTS(reg)));

	/* update this, or overwrite() will be useless! */
	copy_boxp(&cmd->area, (REGION_EXTENTS(pThinc->pScreen, reg)));

	thincBufferCmd(pThinc, pWindow, cmd);

	REGION_UNINIT(pThinc->pScreen, reg);
	copy_reg(raw->reg, bak);
	copy_boxp(&cmd->area, (REGION_EXTENTS(pThinc->pScreen, &raw->reg)));

	thincDbg(T_DBG_CMD, ("%s: finished\n", func));
	stats_end(pThinc, t1, t2, rawStats.exec);
}

#ifdef T_ENABLE_WRITEFILE_FUNC

#define _raw_do_write(fd, ptr, len) \
	err = write(fd, ptr, len); \
	if(err == -1) { \
		return; \
	}

void thincRawWriteToFile(char *data, unsigned int w, 
	unsigned int h, unsigned int Bpp, unsigned int pitch)
{
	static int count = 0;
	unsigned int a, i;
	char fn[255];
	int err, dump; /* fd */
	const char *base = "/tmp/raw-pre/raw";
	register char *curr;
	const int mode = S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH;

	/* Write a header?? */
	const unsigned char write_header = T_TRUE;

	strncpy(fn, base, 16);
	sprintf(fn+16, "%d", count++);

	thincError("file %s\n", fn);

	dump = creat(fn, mode);
	if(dump == -1) {
		return ;
	}

	if(write_header) {
		a = w;
		a = htonl(a);
		_raw_do_write(dump, &a, 4);
		a = h;
		a = htonl(a);
		_raw_do_write(dump, &a, 4);
		a = Bpp;
		a = htonl(a);
		_raw_do_write(dump, &a, 4);
		a= 0;	/* pad */
		_raw_do_write(dump, &a, 4);
	}

	curr = data;

	if(!pitch) pitch = w*Bpp;

	for(i=0; i<h; ++i) {
		_raw_do_write(dump, curr, w*Bpp);
		curr += pitch;
	}

	close(dump);
}
#endif
