/*
 *
 * Header for doubly linked list
 *
 * Author:
 * Ricardo Baratto
 * Network Computing Lab - Columbia University
 */

/* set tabstop = 8 */

/* new list code heavily based on the list implementation in the linux 
 * kernel (linux/list.h) */

#ifndef _THINC_LIST_H_
#define _THINC_LIST_H_

/* 
 * -> I'm not sure inline functions are allowed by [ansi pedantic] so 
 * instead we use macros
 * TODO:
 * -> Add casting to give us some protection in the macros
 */

typedef struct thincList_S {
	struct thincList_S *next, *prev;
#ifdef T_GETSTATS	/* baggage and more bagagge :( */
	unsigned int len;
	unsigned int max_len;
#endif	/* T_GETSTATS */
} thincListRec, *thincListPtr;

/* XXX: don't use this, it produces warnings from -pedantic */
#define t_list_head_init(name) { &(name), &(name) }

/* define a new list */
/* XXX: don't use this, it produces warnings from -pedantic */
#define t_list_head(name) \
	thincListRec name = t_list_head_init(name)

/* init a list */
#define t_list_init(ptr) do { 	\
	(ptr)->next = (ptr); 	\
	(ptr)->prev = (ptr); 	\
	t_list_len_init(ptr);	\
} while(0)

#define t_list_alloc(list) do { \
	list = malloc(sizeof(thincListRec)); \
	if(list) { \
		t_list_init(list); \
	} \
} while(0)

/* 
 * If somebody uses head
 * after invalidated it we have
 * a segfault and lots of fun
 * guaranteed they are to have
 */
#define t_list_inv(head) \
	(head)->next = (head)->prev = NULL;

#define t_list_prev(entry) \
	((entry)->prev)

#define t_list_next(entry) \
	((entry)->next)

/* insert a new entry n between pr and nx. don't use for adding to the tail
 * on empty lists (assignment order matters) */
#define t_list_add(n, pr, nx) do {\
	(nx)->prev = (n); \
	(n)->next = (nx); \
	(n)->prev = (pr); \
	(pr)->next = (n); \
} while(0)

/* add a new entry after the specified head (stacks) */
#define t_list_add_head(n, head) do {		\
	t_list_add(n, head, (head)->next);	\
	t_list_len_inc(head);			\
} while(0)

/* add entry at the tail of the list (queues). can't reuse list_add because
 * we want to be able to use this with empty lists */
#define t_list_add_tail(n, head) do { \
	((head)->prev)->next = (n); \
	(n)->prev = (head)->prev; \
	(n)->next = (head); \
	(head)->prev = (n); \
	t_list_len_inc(head);			\
} while(0)

/* delete entry between pr and nx */
#define t_list_del(pr, nx) do { \
	(nx)->prev = (pr); \
	(pr)->next = (nx); \
} while(0)

#if 0
/* delete entry from list */
#define t_list_del_el(entry) \
	t_list_del((entry)->prev, (entry)->next); \
	t_list_inv(entry); 
#else
/* delete entry and make sure entry is not pointing somewhere crazy */
#define t_list_del_el(entry) do { \
	t_list_del((entry)->prev, (entry)->next); \
	t_list_init(entry); \
} while(0)
#endif

/* checks if the list is empty */
#define t_list_is_empty(list) \
	((list)->next == (list))

/* checks if list has only one element. kind of a hack */
#define t_list_one_el(list) \
	( ( (list)->next != (list) ) && ( (list)->next->next == (list)) )

/* puts list src at the end of list dst. assumes src has something. if
 * you don't know use t_list_move. also src is emptied */
#define t_list_join(dst, src) do { 					\
	thincListPtr __first = (src)->next; 				\
	thincListPtr __last = (src)->prev; 				\
	thincListPtr __tail = (dst)->prev; 				\
									\
	__first->prev = __tail;						\
	__tail->next = __first;						\
	__last->next = (dst); 						\
	(dst)->prev = __last; 						\
	t_list_init(src);						\
	t_list_len_add(dst, t_list_len(src));				\
} while(0)

/* moves src to the end of dst. if src is empty nothing's done */
#define t_list_move(dst, src) \
	if(!t_list_is_empty(src)) { 	\
		t_list_join(dst, src);	\
	}

/* from stddef.h */
#define t_offsetof(type, member) \
	((size_t) &((type*)0)->member)

/* from kernel.h but modified to compile with ansi-pedantic, which doesn't
 * like ({ }) :(
 * but i'm not sure this is completely safe (i'm guessing the initial 
 * definition in the original allows for typechecking?).
 */
#define t_container_of(ptr, type, member) \
	( (type*)( (char*)(ptr) - t_offsetof(type, member) ) )

/* get struct for this entry */
#define t_list_entry(entry, type, member) \
	t_container_of(entry, type, member)

/* get first and last elements of list */
#define t_list_first(head) \
	((head)->next)

#define t_list_last(head) \
	((head)->prev)

#define t_list_first_entry(head, type, member) \
	t_list_entry((head)->next, type, member)

#define t_list_last_entry(head, type, member) \
	t_list_entry((head)->prev, type, member)

/* iterate over a list. pos is the counter (thincListPtr). head is the head
 * of the list */
#define t_list_for_each(pos, n, head) \
	for(pos = (head)->next, n = pos->next; pos != (head); \
		pos = n, n = pos->next)

#define t_list_for_each_fast(pos, head) \
	for(pos = (head)->next; pos != (head); pos = pos->next)

#define t_list_for_each_back(pos, n, head) \
	for(pos = (head)->prev, n = pos->prev; pos != (head); \
		pos = n, n = pos->prev)

/* starts at pos (not pos->next), and goes on until head */
#define t_list_for_each_cont(pos, n, head) \
	for(n = pos->next; pos != (head); pos = n, n = pos->next)

/* iterate over list with a given type. pos is type * to use as a loop
 * counter. n stores the next value, in case pos is deleted within the loop
 * body. head is the head of the list. member is the name of the
 * thincListPtr within the struct */
#define t_list_for_each_entry(pos, n, type, head, member) \
	for(pos = t_list_entry((head)->next, type, member), \
		n = t_list_entry(pos->member.next, type, member); \
 	    &pos->member != (head); \
	    pos = n, n = t_list_entry(n->member.next, type, member) \
	)

#define t_list_for_each_entry_fast(pos, type, head, member) \
	for(pos = t_list_entry((head)->next, type, member); \
 	    &pos->member != (head); \
	    pos = t_list_entry(pos->member.next, type, member) \
	)

#define t_list_for_each_entry_back(pos, n, type, head, member) \
	for(pos = t_list_entry((head)->prev, type, member), \
		n = t_list_entry(pos->member.prev, type, member); \
 	    &pos->member != (head); \
	    pos = n, n = t_list_entry(n->member.prev, type, member) \
	)

/* start at pos and go until the end */
#define t_list_for_each_entry_cont(pos, n, type, head, member) \
	for(n = t_list_entry(pos->member.next, type, member); \
 	    &pos->member != (head); \
	    pos = n, n = t_list_entry(n->member.next, type, member) \
	)

#define t_list_for_each_entry_cont_fast(pos, type, head, member) \
	for( /* empty */ ; \
 	    &pos->member != (head); \
	    pos = t_list_entry(pos->member.next, type, member) \
	)

/* empties a list by repeatedly calling func for each entry in the list
 * and deleting it from the list. the function call will be 
 * func(pThinc, entry)
 */
#define t_list_empty(entry, type, head, member, func, pThinc) do { 	\
	type* __tmp;							\
	t_list_for_each_entry(entry, __tmp, type, head, member) {	\
		t_list_del_el(&(entry)->member);			\
		func (pThinc, entry);					\
	}								\
	t_list_init(head);						\
	t_list_len_set(head, 0);					\
} while(0)

#define t_list_empty_simple(head) do { 					\
	thincListPtr __one, __two;					\
	t_list_for_each(__one, __two, head) {				\
		t_list_del_el(__one);					\
	}								\
	t_list_init(head);						\
	t_list_len_set(head, 0);					\
} while(0)

/* ways to take care of the extra (unwanted) bagagge */
#ifdef T_GETSTATS
#define t_list_len_init(list) \
	(list)->len = (list)->max_len = 0;

#define t_list_len_inc(list) do { 		\
	if(++(list)->len > (list)->max_len)	\
		(list)->max_len = (list)->len;	\
} while(0)

#define t_list_len_dec(list) --(list)->len

#define t_list_len_add(list, l) do {		\
	(list)->len += (l);			\
	if((list)->len > (list)->max_len)	\
		(list)->max_len = (list)->len;	\
} while(0)

#define t_list_len_sub(list, l) (list)->len -= (l)

#define t_list_len(list) (list)->len
#define t_list_maxlen(list) (list)->max_len

#define t_list_len_set(list, l) (list)->len = (l)
#define t_list_maxlen_set(list, l) (list)->max_len = (l)

#else /* T_GETSTATS */
#define t_list_len_init(list)

#define t_list_len_inc(list)
#define t_list_len_dec(list) 
#define t_list_len_add(list, l)
#define t_list_len_sub(list, l)

#define t_list_len(list)
#define t_list_maxlen(list)

#define t_list_len_set(list, l)
#define t_list_maxlen_set(list, l)

#endif	/* T_GETSTATS */

#endif	/* _THINC_LIST_H_ */
