/* 
 *
 * Header for offscreen pixmap operations
 *
 * Author:
 * Ricardo Baratto
 * Network Computing Lab - Columbia University
 */

/* set tabstop = 8 */
#ifndef _THINC_PIXMAP_H_
#define _THINC_PIXMAP_H_

/* How it works: (see also thincCommand.h)
 * Each Pixmap has a private structure, thincPixmapRec which contains
 * the list of drawing commands queued in that Pixmap. Each command
 * is represented by a thincCmdRec structure (see below).
 * The list is a linked list of pointers to this structures, like:
 * ---------   ---------         ---------
 * | Cmd1  |-->| Cmd 2 |-->...-->| Cmd n | 
 * |   *   |<--|   *   |<--   <--|   *   |
 * ----|----   ----|----         ----|----
 *     |           |                 |     ------------------
 *     |           |                 |--->|CmdRec | cmd data |
 *     |           |                       ------------------
 *     |           |    ------------------
 *     |           |-->|CmdRec | cmd data |
 *     |                ------------------
 *     |     -------------------
 *     |--->| CmdRec | cmd data |
 *           -------------------
 *
 * The cmd data in each command's box is dependent to each type of command.
 * There are 4 possible operations that can be done on each command:
 * - Create(): Receives parameters for the command and returns a pointer
 *   to a CmdRec structure (CmdPtr from now on) which can be added to
 *   the pixmap's queue.
 * - Destroy(): Receives a CmdPtr and frees up all the memory used by it.
 *   After Destroy is finished, the passed pointer is no longer valid.
 * - Copy(): Receives a CmdPtr and returns a pointer to a copy of the
 *   passed command. It should perform a deep copy.
 * - Modify(): Receives a CmdPtr and modifies it according to the
 *   parameters passed. To simplify the code, the modification will always
 *   be either translate, clip or both. The function prototype reflects
 *   this. It should return a value, informing the caller if as a result
 *   of the modification, the command doesn't affect any area in
 *   the pixmap. In this case, Destroy() should be called and the
 *   command taken out of the list.
 * - Execute(): Receives a CmdPtr and makes it happen on the screen.
 *   Should call appropiate function to send the command to the
 *   client(s).
 *   
 * Upon startup, we fill up a table with pointers to the functions for
 * each operation for each command.
 *
 *************
 * 
 * => When a command needs to be queued the Create() function will
 * be called.
 * => When a pixmap is destroyed the Destroy() function will be called
 * on each command in its queue.
 * => When a CopyArea is made from one pixmap to another pixmap, the
 * Copy() and Modify() functions will be called. If as a result of
 * Modify() (e.g. clipping) the command doesn't affect any area
 * in the pixmap, Destroy() will be called.
 *
 *************
 * New commands are always added to the end of the list. If they
 * overwrite any of the commands already in the queue, that command
 * should be modified appropiately (could be taken off the queue)
 */

/* Pixmap Private */
typedef struct thincPixmap_S {
	ulong	id;		/* the pixmap's id. Have our own instead of
				   relying on X's */
	thincListRec list;	/* the list of commands */
#ifdef T_USERENDER
	int	i_got_it:1;	/* used by the different layers in the 
				   RENDER code to tell Composite (the
				   lowest level function) that it shouldn't
				   try to do anything about this update */
	int	pad:31;		/* for the future? */
#endif
} thincPixmapRec, *thincPixmapPtr;

/* some macros to access members of the private struct from the Pixmap */

/* get a pixmap's private id. beautiful, eh? :) */
#define t_pixID(pPixmap) \
	((thincPixmapPtr)(THINCPIXPtr((pPixmap))))->id

#ifdef T_USERENDER
#define t_got_it(pPixmap) \
	( ((thincPixmapPtr)(THINCPIXPtr((pPixmap))))->i_got_it)

#define t_get_it(pPixmap) \
	((thincPixmapPtr)(THINCPIXPtr((pPixmap))))->i_got_it = 1

#define t_let_it_go(pPixmap) \
	((thincPixmapPtr)(THINCPIXPtr((pPixmap))))->i_got_it = 0
#endif

/* Return a pointer to the pixmap's data. 
 * FIXME: Hackish, since the way pixmaps are
 * stored is implementation dependent. However, current video drivers all
 * use the fb code which in turn stores the pixmap data at this location :)
 * And since we have it isolated here, if we find a better way to access it 
 * or the implementation changes we won't cry.. too much :)
 * p is a PixmapPtr */
#ifdef T_WIN32
#define thinc_getPixmap(p) \
	(p)->data
#else /* T_WIN32 */
#define thinc_getPixmap(p) \
	(p)->devPrivate.ptr
#endif /* T_WIN32 */

/********* Functions ********/

/* Functions to manipulate the PixmapPtr devPrivate structure */
void thincInitPixmapPriv(PixmapPtr pPix);
void thincUninitPixmapPriv(PixmapPtr pPix);

/* The boss functions. Call everything else */
int thincQueueCommand(thincScreenPtr pThinc, PixmapPtr pPix, int type,
	RegionPtr reg, ...);
int thincQueueThisCommand(thincScreenPtr pThinc, thincPixmapPtr pPriv,
	thincCmdPtr cmd, RegionPtr reg);
int thincCopyPixmapQueue(thincScreenPtr pThinc, PixmapPtr srcpix, 
	PixmapPtr dstpix, int srcx, int srcy, int w, int h, int dstx, 
	int dsty);
int thincExecPixmapQueue(thincScreenPtr pThinc, WindowPtr dst, PixmapPtr pix, 
	int srcx, int srcy, int w, int h, int dstx, int dsty, 
	RegionPtr dstReg);
void thincEmptyPixmapQueue(PixmapPtr pPix);

#endif /* _THINC_PIXMAP_H_ */
