/* Copyright 2003, 2004, 2005 Ricardo Baratto */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
 *
 * Command scheduling functions
 *
 * Author:
 * Ricardo Baratto
 * Network Computing Lab - Columbia University
 *
 */

/* 
 * Structures defined in thincCommand.h
 * How it works:
 * Scheduler structure has a series of queues each corresponding to the size 
 * of the commands in it. Size ranges are based on power of 2, starting at
 * 2K. When a command needs to be buffered for a client, we figure out
 * which queue it belongs to and then add it there. Order inside the queue
 * doesn't matter much (for now). At flush time, the queues are flushed
 * in order according to their size.
 *
 * Note that the standard client command buffer still exists, and the queues 
 * and the buffer share the pointers to the actual commands.
 */

/*
 * ****** Dealing with dependencies ***********
 * scenarios:
 * 1. a command explicitly depends on others that draw before it
 * 2. a command implicitly makes other commands depend on it because
 *   it doesn't allow itself to be partially overwritten
 *
 *   right now #1 is due to transparency, and it only affects glyphs. in
 *   this case, we have a new flag, T_CMD_DEPS_NEEDS, which tells us
 *   that before buffering and scheduling it, we need to figure out
 *   which commands it depends on, and make sure we send it _after_ that
 *   one. many ways to do this, but we chose the simple and hackish way:
 *   figure out dependencies, find the largest size, and have the
 *   command return that size to the scheduler. voil? *
 *   #2 has the potential to be a little bit more complicated. right now
 *   is for performance reasons and it only affects solid and pixmap
 *   fills (but really just pixmap fills). In this case, we cannot allow
 *   commands which _think_ they overwrote a fill to be scheduled ahead
 *   of the fill. the solution follows our current simple and hackish
 *   trend. we make all pixmap and solid fills always go to the first
 *   queue! to avoid problems where the pixmap fill may be to big and
 *   screw up something else, we also reduced the maximum size a tile
 *   may have to be considered a pixmap fill.  from 128x128 to 48x48.
 *   48x48 in 24bpp comes out to about 10KB which shouldn't be _that_
 *   bad.
 */

/* set tabstop = 8 */

#include "thinc.h"
#ifndef T_WIN32
#include <sys/poll.h>
#endif

static __inline__ int thincGetQueue(unsigned int size);
static int thincSchedCommand(thincClientPtr cl, thincCmdPtr cmd);
static int thincCmdIsRT(thincClientPtr cl, thincSchedPtr sched,
	thincCmdPtr cmd); 
static int thincSchedCommandRT(thincClientPtr cl, thincSchedPtr sched,
	thincCmdPtr cmd);
static int thincSchedCommandNormal(thincClientPtr cl, thincSchedPtr sched,
	thincCmdPtr cmd);
static int thincCmdSetSchedSize(thincScreenPtr pThinc, thincCmdPtr cmd,
	unsigned int size);

/* entry point to scheduling a command */
int thincAddCommand(thincClientPtr cl, thincCmdPtr cmd) 
{
	t_dbg_name(func)
	int err;

	if(cl->parent->doSched) {
		if(cmd->flags & T_CMD_DEPS_NEEDS) {
			thincDbg(T_DBG_CLIENT, ("%s: client may have "
				"dependencies, looking for 'em\n", func));
			err = thincCmdFindDeps(cl->parent, cmd, &cl->cmdBuffer);
			if(err == T_ERROR) {
				thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't "
					"find dependencies for command "
					"%#lx (type %d)\n", func, cmd->id,
					cmd->type));
				return T_ERROR;
			}
		}
		err = thincSchedCommand(cl, cmd);
		if(err == T_ERROR)
			return T_ERROR;
	}
	err = thincAddCmdList(&cl->cmdBuffer, cmd);
	return err;
}

static int thincSchedCommand(thincClientPtr cl, thincCmdPtr cmd)
{
	thincSchedPtr sched = cl->sched;
	if(!sched) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Scheduler has not been "
			"initialized\n", __func__));
		return T_ERROR;
	}
	if(thincCmdIsRT(cl, sched, cmd))
		return thincSchedCommandRT(cl, sched, cmd);
	else
		return thincSchedCommandNormal(cl, sched, cmd);
}

/* since there's only one rt queue, just put it there */
static int thincSchedCommandRT(thincClientPtr cl, thincSchedPtr sched,
	thincCmdPtr cmd) 
{
	t_dbg_name(func)
	unsigned int qnum;
	thincListPtr queue;
	thincDbg(T_DBG_CLIENT, ("%s: queueing high priority command %#lx\n", 
		func, cmd->id));
	
	qnum = T_SCHED_RT_IDX;
	queue = &sched->queues[qnum];
	if(thincAddCmdSched(queue, cmd) == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't add command "
			"to command queue %d\n", func, qnum));
		return T_ERROR;
	}

	thincDbg(T_DBG_CLIENT, ("%s: finished\n", func));
	return T_OK;
}

static int thincSchedCommandNormal(thincClientPtr cl, thincSchedPtr sched,
	thincCmdPtr cmd)
{
	t_dbg_name(func)
	unsigned int size, qnum;
	thincListPtr queue;
	thincScreenPtr pThinc = cl->parent;
	
	thincDbg(T_DBG_CLIENT, ("%s: adding normal command %#lx to client's "
		"queue\n", func, cmd->id));

	size = thincCmdGetSize(pThinc, cmd);
	if(size >= T_SCHED_MAXSIZE) 
		qnum = T_SCHED_NORMALQUEUES-1;
	else
		qnum = thincGetQueue(size);

	qnum += T_SCHED_RTQUEUES;

	thincDbg(T_DBG_CLIENT, ("%s: command's size %d bytes, queue %d\n", 
		func, size, qnum));

	queue = &sched->queues[qnum];

	if(thincAddCmdSched(queue, cmd) == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't add command "
			"to scheduler queue %d\n", func, qnum));
		return T_ERROR;
	}

	thincDbg(T_DBG_CLIENT, ("%s: finished\n", func));
	return T_OK;
}

int thincFlushScheduler(thincScreenPtr pThinc, thincClientPtr cl)
{
	t_dbg_name(func)
	register int i, err;
	register thincListPtr queue;
	thincSchedPtr sched = cl->sched;

	thincDbg(T_DBG_ANNOY, ("%s: flushing scheduler queues\n", func));

	t_bug_on_2(!cl->sched,
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Scheduler has not been "
			"initialized for this client\n", func)),
		return T_ERROR
	);

	err = thincDoPoll(cl->sock, POLLOUT, 0);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't flush client's "
			"buffer\n", func));
		return T_ERROR;
	} else if(err == T_TIMEOUT) {
		/* can't write to this client now */
		thincDbg(T_DBG_ANNOY, ("%s: Client is busy, come back "
			"later. finished\n", func));
		return T_PART;
	}

	for(i=0; i<T_SCHED_NUMQUEUES; ++i) {
		queue = &sched->queues[i];
		if(t_list_is_empty(queue) ) {
			thincDbg(T_DBG_ANNOY, ("%s: queue %d empty\n", func,
				i));
			continue;
		}

		thincDbg(T_DBG_ANNOY, ("%s: flushing queue %d\n", func, i));

		stats_dbg(("Flush:Sched[%d]:Client[%s]: len %u "
			"max_len %u\n", i, cl->id, 
			t_list_len(queue), t_list_maxlen(queue))
		);

		err = thincFlushListSched(pThinc, cl, queue);
		if(err != T_OK) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't flush "
				"queue %d\n", func, i));
			return T_ERROR;
		}
	}

	thincDbg(T_DBG_ANNOY, ("%s: finished\n", func));
	return T_OK;
}

/* returns the queue to which a command of size size belongs.
 * TODO: Is there a better way to do this? */
static __inline__ int thincGetQueue(unsigned int size)
{
	int count = 0;
	size >>=T_SCHED_FIRST;
	while(size) {
		++count;
		size >>=1;
	}
	return count;
}

/* initialize scheduler */
int thincSchedInit(thincClientPtr cl)
{
	int i;
	thincSchedPtr sched;

	sched = malloc(sizeof(thincSchedRec));
	if(!sched) {
		cl->sched = NULL;
		return T_ERROR;
	}
	for(i=0; i<T_SCHED_NUMQUEUES; ++i) {
		t_list_init(&sched->queues[i]);
	}
	cl->sched = sched;
	return T_OK;
}

void thincSchedDestroy(thincClientPtr cl)
{
	int i;
	thincSchedPtr sched = cl->sched;
	
	if(!sched)
		return;
	for(i=0; i<T_SCHED_NUMQUEUES; ++i) {
		stats_log(("Close:Sched[%s]:Client[%d]: len %u max_len %u\n",
			cl->id, i, t_list_len(&sched->queues[i]),
			t_list_maxlen(&sched->queues[i])));
		if(!t_list_is_empty(&sched->queues[i])) {
			thincEmptyCmdListSched(cl->parent, &sched->queues[i]);
		}
	}
	free(sched);
	cl->sched = NULL;
}

static int thincCmdIsRT(thincClientPtr cl, thincSchedPtr sched, 
	thincCmdPtr cmd)
{
	return (cmd->flags & T_CMD_RT ? TRUE: FALSE);
}

int thincShouldYield(thincClientPtr cl)
{
	unsigned long now;
	int err;

	if(!cl->parent->doSched)
		return T_NO;
	
	gettime_ms(now);
	if(now - cl->last < T_RESPONSE) {
		cl->last = now;
		return T_NO;
	}

	cl->last = now;

	/* ok, we have waited long enough. check if any client has something
	 * to say */
	err = thincClientsHaveEvents(cl->parent);
	if(err == T_ERROR || err == T_YES)
		return err;

	/* ok, nothing there, check if we can keep on writing */
	err = thincDoPoll(cl->sock, POLLOUT, 0);
	if(err == T_ERROR) {
		return T_ERROR;
	} else if(err == T_PART) {
		return T_YES;
	}

	return T_NO;
}

/* returns T_YES if any connected clients has events pending */
int thincClientsHaveEvents(thincScreenPtr pThinc)
{
	int err;
	thincClientPtr cl;

	t_list_for_each_entry_fast(cl, thincClientRec, &pThinc->clients, link) {
		err = thincDoPoll(cl->sock, POLLIN, 0);
		if(err == T_ERROR)
			return T_ERROR;
		else if(err == T_OK)
			return T_YES;
		thincDbg(T_DBG_ANNOY, ("%s: client %s inactive\n", __func__,
			cl->id));
	}

	return T_NO;
}

/* Find the commands cmd depends on (those with which it overlaps). From
 * those, find the one with the largest size and make the command use that
 * size (or something larger) as its size for the scheduler. this way 
 * we guarantee that the command won't be sent to the client before all 
 * of its dependencies. hackish but simple and it works :) */
int thincCmdFindDeps(thincScreenPtr pThinc, thincCmdPtr cmd, thincListPtr list)
{
	t_dbg_name(func)

	thincCmdPtr curr, next;
	BoxPtr currArea;
	int inreg;
	RegionPtr pReg;
	unsigned int max_sz=0, curr_sz;

	thincDbg(T_DBG_CLIENT, ("%s: Looking for dependencies of command %#lx "
		"(type %d)\n", func, cmd->id, cmd->type));
	stats_inc(pThinc, find_deps);

	thincCmdGetRegion(pThinc, cmd, &pReg);
	print_reg(T_DBG_CLIENT, pThinc->pScreen, pReg, "overlap region:");

	t_list_for_each_entry(curr, next, thincCmdRec, list, list) {
		thincDbg(T_DBG_CLIENT, ("%s: id %#lx, type %d\n",
			func, curr->id, curr->type));

		currArea = &curr->area;
		/*
		 * this checks if currArea is inside, outside or miti-miti
		 * of pReg.
		 * then:
		 * -> if currArea is inside pReg, then the command is
		 *  completely clipped out and we should just delete it.
		 * -> if currArea is outside pReg, then we don't have
		 *  to do anything with this dud so just go to next
		 *  command.
		 * -> if there's some overlap but no clear winner, then
		 *  need to clip the command with currReg - pReg
		 */
		inreg = RECT_IN_REGION(pThinc->pScreen, pReg, currArea);
		if(inreg == rgnOUT) {
			/* sweet, next */
			thincDbg(T_DBG_CLIENT, ("%s: our command does not "
				"depend on this command, next svp\n", func));
			continue;
		}
		/* overlap. add it to the dependency list */
		curr_sz = thincCmdGetSize(pThinc, curr);
		thincDbg(T_DBG_CLIENT, ("%s: found dependency on command %#lx "
			"(type %d). it has size %d bytes\n", func, curr->id, 
			curr->type, curr_sz));
		if(curr_sz > max_sz) {
			max_sz = curr_sz;
			thincDbg(T_DBG_CLIENT, ("%s: new max size %d bytes\n",
				func, max_sz));
		}
	}

	if(max_sz) {
		thincDbg(T_DBG_CLIENT, ("%s: setting command's sched size to "
			"%d bytes\n", func, max_sz));
		if(thincCmdSetSchedSize(pThinc, cmd, max_sz) == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't set command's "
				"scheduler size\n", func));
			return T_ERROR;
		}
	}

	thincDbg(T_DBG_CLIENT, ("%s: finished\n", func));
	return T_OK;
}

static int thincCmdSetSchedSize(thincScreenPtr pThinc, thincCmdPtr cmd,
	unsigned int size)
{
	if(thincModifyCommand(pThinc, cmd, T_CMD_SCHEDSZ, size) == 
		(thincCmdPtr) T_ERROR) 
		return T_ERROR;
	else
		return T_OK;
}
