/* Copyright 2003, 2004, 2005 Ricardo Baratto */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
 * $Id; $
 *
 * SSL server functions
 *
 * Author:
 * Angelos Stavrou <angel@cs.columbia.edu>
 * Stelianos Sidiroglou-Douskos <stelios@cs.columbia.edu>
 * Network Security Lab - Columbia University
 *
 * Modified for THINC:
 * Ricardo Baratto
 * Network Computing Lab - Columbia University
 */

/* set tabstop = 8 */

#include "thinc_ssl.h"

static int thinc_ssl_load_certificates(SSL_CTX* ctx, const char* cert, 
	const char* key);
static SSL_CTX* thinc_ssl_init_server_ctx();

/* initialize */
SSL_CTX* thinc_ssl_init_server(const char *cert_location, 
	const char *key_location)
{
	SSL_CTX *ctx;
	int err;

	/* init context */
	ctx = thinc_ssl_init_server_ctx();
	if(!ctx)
		return NULL;

	/* initialize SSL */
#if 1
	err = thinc_ssl_load_certificates(ctx, cert_location, key_location);
	if(err == T_SSL_ERROR) {
		SSL_CTX_free(ctx);
		return NULL;
	}
#endif

	return ctx;
}

int thinc_ssl_accept(int client, SSL_CTX *ctx, SSL **ssl_p)
{
	SSL *ssl;
	int err, ssl_err;

	if(*ssl_p) {
		ssl = *ssl_p;
	} else {
		ssl = SSL_new(ctx);      /* get new SSL state with context */
		if(!ssl) {
			return T_SSL_ERROR;
		}
	}

	SSL_set_fd(ssl, client);  /* set connection socket to SSL state */

	/* TODO Important TODO
	 * If the underlying BIO is non-blocking, SSL_accept() will also 
	 * return when the underlying BIO could not satisfy the needs of
	 * SSL_accept() to continue the handshake, indicating the problem by
	 * the return value -1. In this case a call to SSL_get_error() with the
	 * return value of SSL_accept() will yield SSL_ERROR_WANT_READ or
	 * SSL_ERROR_WANT_WRITE. The calling process then must repeat the call
	 * after taking appropriate action to satisfy the needs of
	 * SSL_accept().  The action depends on the underlying BIO. When using
	 * a non-blocking socket, nothing is to be done, but select() can be
	 * used to check for the required condition.
	 */
	err = SSL_accept(ssl);
	if(err == 0 ) {
		/* connection closed */
		SSL_free(ssl);
		*ssl_p = NULL;
		return T_SSL_CLOSED;
	} else if(err < 0) {
		/* check what kind of error */
		ssl_err = SSL_get_error(ssl, err);
		if( 	(ssl_err == SSL_ERROR_WANT_READ) || 
			(ssl_err == SSL_ERROR_WANT_WRITE) ) {
			*ssl_p = ssl;
			return T_SSL_WOULDBLOCK;
		} else {
			SSL_free(ssl);
			*ssl_p = NULL;
			return T_SSL_ERROR;
		}
	}
	
	/* XXX: Debug. Print cipher being used XXX */
#if 0
	{
		FILE *f = fopen("/tmp/cipher", "w");
		fprintf(f, "Cipher is %s\n", SSL_get_cipher_name(ssl));
		fclose(f);
	}
#endif
	
	*ssl_p = ssl;
	return T_SSL_OK;
}

int thinc_ssl_write(SSL *ssl, const void *buf, int count)
{
	int err, ssl_err;

	err = SSL_write(ssl, buf, count);
	if(err == 0) {
		return T_SSL_CLOSED;
	} else if(err < 0) {
		ssl_err = SSL_get_error(ssl, err);
		/* man page says both errors can be returned since at any
		 * point in time, a handshake may be necessary... */
		if( 	(ssl_err == SSL_ERROR_WANT_READ) || 
			(ssl_err == SSL_ERROR_WANT_WRITE) ) {
			return T_SSL_WOULDBLOCK;
		} else {
			return T_SSL_ERROR;
		}
	}

	return err;
}

static SSL_CTX* thinc_ssl_init_server_ctx()
{
	SSL_METHOD *method;
	SSL_CTX *ctx;

	OpenSSL_add_all_algorithms();	/* load & register all cryptos, etc. */
	SSL_load_error_strings();	/* load all error messages */
	/* create new server-method instance */
	method = SSLv23_method();
	ctx = SSL_CTX_new(method);	/* create new context from method */
	if(!ctx) {
		ERR_print_errors_fp(stderr);
		return NULL;
	}
	/* TODO: What's up with this path?? Should check return values?? */
#if 0
	SSL_CTX_load_verify_locations(ctx,"../client/cacert.pem", NULL);
	SSL_CTX_set_verify(ctx, SSL_VERIFY_PEER, NULL);
#endif

	return ctx;
}

static int thinc_ssl_load_certificates(SSL_CTX* ctx, const char* cert, 
	const char* key)
{
	int err; 

	/* set the local certificate from CertFile */
	err = SSL_CTX_use_certificate_file(ctx, cert, SSL_FILETYPE_PEM);
	if(err <= 0 ) {
	    return T_SSL_ERROR;
	}

	/* set the private key from KeyFile (may be the same as CertFile) */
	err = SSL_CTX_use_PrivateKey_file(ctx, key, SSL_FILETYPE_PEM);
	if(err <= 0 ) {
	    return T_SSL_ERROR;
	}

	/* verify private key */
	if( !SSL_CTX_check_private_key(ctx) ) {
		/* Private key does not match the public certificate */
		return T_SSL_ERROR;
	}

	return T_SSL_OK;
}
