/* Copyright 2004, 2005 Lei Zhang, Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/******************************Module*Header*******************************\
*
*                       **************************
*                       * THINC Related Routines *
*                       **************************
*
* Module Name: thinc.c
*
* This module contains the functions that pushes intermediate commands into
* the command channel. 
*
***************************************************************************/

#include "driver.h"
#include "string.h"
#include "common.h"
#include "MirrCmdProto.h"

#define RAWUPDAT_DBG_LEV 10
#define COPYBITS_DBG_LEV 2

//
// Push an intermedia RAW_UPDATE command into the command channel
//

INT RawUpdate(SURFOBJ*   psoDst,
              RECTL*	 rct )
{
	PMirrSurf	psurfDst = (PMirrSurf)(psoDst->dhsurf);
		
	int        i, j, datasize, pixelbits, screenstride, stride, height, width, offset;
			
	PVOID      data, pscan;

	PThincCmd  pCmd;

	pixelbits = psurfDst->ulBitCount;

	width = rct->right - rct->left;
	height = rct->bottom - rct->top;

	// heuristic
	// if the raw update region include the entire surface and is not too big,
	// it might be used fox pixmap fill. 
	//if (width == psurfDst->cx && height == psurfDst->cy &&
	//	width * height <= 2000)
#if 0
	if (height == 1)
	{
		PixmapFill(psurfDst, rct->left, rct->top, width, height, 1, rct);
		return 0;
	}
#endif
	

	datasize = THINC_HEADERSIZE + T_FB_RUPDATE_SIZE/* + height * width * pixelbits / 8*/;

	pCmd = thincCmdBlockAlloc(&(psurfDst->ppdev->pCmdBlock), datasize, &offset);
	
	if (pCmd!= NULL)
	{
		PVOID				data = pCmd->cmdData;
		MirrCmdHeader*		header = (MirrCmdHeader*) pCmd->cmdData;
		thinc_rawUpdate*	rupdate = (thinc_rawUpdate*)((char*)data + THINC_HEADERSIZE);

		header->type = T_FB_RUPDATE;
		header->dstId = psurfDst->id;
		header->srcId = NULL_SURF_ID;

		rupdate->x = rct->left;
		rupdate->y = rct->top;
		rupdate->width = width;
		rupdate->height = height;

		DISPDBG((1, "T_FB_RUPDATE: dstId= %d, [(%d, %d), (%d, %d)]\n", header->dstId, rupdate->x, rupdate->y, 
				rupdate->x + rupdate->width, rupdate->y + rupdate->height));

#ifdef TIME_STAMP
		// Time stamp the command
		psurfDst->ppdev->pfQuerySystemTime(&pCmd->timeStamp[0]);
#endif

		if (!thincPushCmdWithOffset(psurfDst->ppdev, psurfDst->ppdev->pCmdBlock, offset))
		{
			DISPDBG((RAWUPDAT_DBG_LEV, "  RawUpdate: pushing command failed\n")); 
		}					   
		
	}
	else
	{
		DISPDBG((RAWUPDAT_DBG_LEV, " RawUpdate: command data allocation failed\n"));
	}

	return 0;
}

//
// Push a fake RAW_UPDATE intermediate command into the command channel.
// This routine is used only when the user mode part of the server starts.
//

INT FakeRawUpdate(PMirrSurf	psurfDst)
{
	int        i, j, datasize, pixelbits, screenstride, stride, height, width, offset;
			
	PVOID      data, pscan;

	PThincCmd  pCmd;

	pixelbits = psurfDst->ulBitCount;

	width = psurfDst->cx;
	height = psurfDst->cy;
		
	width = (width == 0)? 1 : width;
	height = (height == 0)? 1 : height;

	datasize = THINC_HEADERSIZE + T_FB_RUPDATE_SIZE /*+ height * width * pixelbits / 8*/;

	pCmd = thincCmdBlockAlloc(&(psurfDst->ppdev->pCmdBlock), datasize, &offset);
	
	if (pCmd!= NULL)
	{
		PVOID				data = pCmd->cmdData;
		MirrCmdHeader*		header = (MirrCmdHeader*) pCmd->cmdData;
		thinc_rawUpdate*	rupdate = (thinc_rawUpdate*)((char*)data + THINC_HEADERSIZE);
		//PVOID				pixels = (PVOID)((PCHAR)data + THINC_HEADERSIZE + T_FB_RUPDATE_SIZE);

		header->type = T_FB_RUPDATE;
		header->dstId = psurfDst->id;
		header->srcId = NULL_SURF_ID;

		rupdate->x = 0;
		rupdate->y = 0;
		rupdate->width = width;
		rupdate->height = height;
		
		DISPDBG((RAWUPDAT_DBG_LEV, " RawUpdate: command data allocation succeed\n"));

#ifdef TIME_STAMP
		// Time stamp the command
		psurfDst->ppdev->pfQuerySystemTime(&pCmd->timeStamp[0]);
#endif

		if (!thincPushCmdWithOffset(psurfDst->ppdev, psurfDst->ppdev->pCmdBlock, offset))
		{
			DISPDBG((RAWUPDAT_DBG_LEV, "  FakeRawUpdate: pushing command failed\n")); 
		}					   
		
	}
	else
	{
		DISPDBG((RAWUPDAT_DBG_LEV, " FakeRawUpdate: command data allocation failed\n"));
	}

	return 0;
}

//
// Push a SOLID_FILL intermediate command into the command channel
//

INT SolidFill(PMirrSurf	 psurfDst,
			  ULONG		 iSolidColor,
			  LONG       lNumRects,
              RECTL*	 pRects )
{
	PPDEV		ppdev = psurfDst->ppdev;

	int			i, datasize, offset;

	RECTL*		prct;
	PThincCmd	pCmd;
	PVOID		data;
	MirrCmdHeader* header;
	thinc_fbSFill* fbsfill;

	thinc_rectRegion* ptRects, *ptrct;

	DISPDBG((COPYBITS_DBG_LEV, "doSFill == true, lNumRects = %d\n", lNumRects));

	datasize = THINC_HEADERSIZE + T_FB_SFILL_SIZE + lNumRects * T_RECTREGION_SIZE;

	pCmd = thincCmdBlockAlloc(&(ppdev->pCmdBlock), datasize, &offset);

	if (pCmd != NULL) 
	{
		data = pCmd->cmdData;
		header = (MirrCmdHeader*)data;
		header->type = T_FB_SFILL;
		header->dstId = psurfDst->id;
		header->srcId = NULL_SURF_ID;

		DISPDBG((COPYBITS_DBG_LEV, " DrvBitBlt: data allocation succeed. data_size = %d\n", datasize));

		fbsfill = (thinc_fbSFill*)((char*)data + THINC_HEADERSIZE);
		fbsfill->numrects = lNumRects;
		fbsfill->pixel = iSolidColor;

		ptRects = (thinc_rectRegion*)((char*)data + THINC_HEADERSIZE + T_FB_SFILL_SIZE);
		for (i = 0; i < lNumRects; i++)
		{
			prct = &pRects[i];
			ptrct = &ptRects[i];
			ptrct->x = prct->left;
			ptrct->y = prct->top;
			ptrct->height = prct->bottom - prct->top;
			ptrct->width = prct->right - prct->left;
		}
#ifdef TIME_STAMP
		// Time stamp the command
		ppdev->pfQuerySystemTime(&pCmd->timeStamp[0]);
#endif

		if (!thincPushCmdWithOffset(ppdev, ppdev->pCmdBlock, offset))
		{
			DISPDBG((1, "  DrvBitBlt -- pushing command failed\n")); 
			EngFreeMem(data);
		}	
	}
	return 0;
}

//
// Push a PIXMAP_FILL intermediate command into the command channel
//

INT PixmapFill(PMirrSurf psurfDst,
			   INT       left,
			   INT       top,
			   INT       width,
			   INT       height,
			   LONG      lNumRects,
               RECTL*	 pRects )
{
	PPDEV		ppdev = psurfDst->ppdev;

	int			i, datasize, offset, pixmapSize;

	RECTL*		prct;
	PThincCmd	pCmd;
	PVOID		data;
	char*		pScanSurf;
	char*		pScanPixmap;
	int			stride;

	MirrCmdHeader* header;
	thinc_fbPFill* fbpfill;
	thinc_rectRegion* ptRects, *ptrct;

	pixmapSize = width * height * (psurfDst->ulBitCount >> 3);
	datasize = THINC_HEADERSIZE + T_FB_PFILL_SIZE + pixmapSize + lNumRects * T_RECTREGION_SIZE;

	pCmd = thincCmdBlockAlloc(&(ppdev->pCmdBlock), datasize, &offset);

	if (pCmd != NULL) 
	{
		data = pCmd->cmdData;
		header = (MirrCmdHeader*)data;
		header->type = T_FB_PFILL;
		header->dstId = psurfDst->id;
		header->srcId = NULL_SURF_ID;

		DISPDBG((COPYBITS_DBG_LEV, " DrvBitBlt: data allocation succeed. data_size = %d\n", datasize));

		fbpfill = (thinc_fbPFill*)((char*)data + THINC_HEADERSIZE);
		fbpfill->numrects = lNumRects;
		fbpfill->x = left;
		fbpfill->y = top;
		fbpfill->width = width;
		fbpfill->height = height;

		pScanSurf = ((char*)psurfDst->pscan) 
			+ (psurfDst->lDelta * top) + left * (psurfDst->ulBitCount >> 3);
		pScanPixmap = ((char*)data + THINC_HEADERSIZE + T_FB_PFILL_SIZE);
		stride = width * (psurfDst->ulBitCount >> 3);

		for (i = 0; i < height; i++)
		{
			memcpy(pScanPixmap, pScanSurf, stride);
			pScanPixmap += stride;
			pScanSurf += psurfDst->lDelta;
		}

		ptRects = (thinc_rectRegion*)((char*)data + THINC_HEADERSIZE + T_FB_PFILL_SIZE + pixmapSize);
		for (i = 0; i < lNumRects; i++)
		{
			prct = &pRects[i];
			ptrct = &ptRects[i];
			ptrct->x = prct->left;
			ptrct->y = prct->top;
			ptrct->height = prct->bottom - prct->top;
			ptrct->width = prct->right - prct->left;
		}
#ifdef TIME_STAMP
		// Time stamp the command
		ppdev->pfQuerySystemTime(&pCmd->timeStamp[0]);
#endif

		if (!thincPushCmdWithOffset(ppdev, ppdev->pCmdBlock, offset))
		{
			DISPDBG((1, "  DrvBitBlt -- pushing command failed\n")); 
			EngFreeMem(data);
		}	
	}
	return 0;
}

//
// Push a GLYPH intermediate command into the command channel
//
BOOL Glyph(PDEV*	ppdev,
		   SURFOBJ* psoDst,
		   RECTL*	pRect,
		   ULONG	ulForeColor)
{
	int				i, j, datasize, bitmapSize, width, height, bitmapwidth, offset;
	char*			bitmap;
	PVOID			data;
	PThincCmd		pCmd;
	MirrCmdHeader*	header;
	thinc_fbGlyph*	fbglyph;
	PMirrSurf	psurfDst = (PMirrSurf)(psoDst->dhsurf);

	width = pRect->right - pRect->left;
	height = pRect->bottom - pRect->top;
	
	width = (width == 0) ? 1 : width;
	height = (height == 0) ? 1 : height;
	
	bitmapwidth = (width + 7) >> 3;

	bitmapSize = bitmapwidth * height;

	datasize = THINC_HEADERSIZE
				+ T_FB_GLYPH_SIZE 
				+ bitmapSize;

	pCmd = thincCmdBlockAlloc(&(ppdev->pCmdBlock), datasize, &offset);
    
	if (pCmd != NULL) 
	{
		PMirrSurf	psurfDst = (PMirrSurf)(psoDst->dhsurf);
		INT			pixelbits = psurfDst->ulBitCount;
		INT			screenstride;
		PVOID		pscan;
		INT			alignedWidth = width >> 3;
		INT			remainingBits = width & 0x07;
		INT			remainingShift = 8 - remainingBits;

		//
		// FIXME: We use ULONG because we currently deal with the 32-bit pixels only.
		//
		ULONG*		pulPixel;	
		ULONG*		pulLine;
		char		cBitsByte, bit;

		screenstride = psoDst->lDelta;

		if (screenstride > 0 || ((psoDst->fjBitmap) & BMF_TOPDOWN))
		{
			pscan = (PCHAR)psoDst->pvScan0 + pRect->top * screenstride
				+ (pRect->left * pixelbits >> 3);
		}
		else
		{
			//
			// I believe it rarely comes here.
			//
			pscan = (PCHAR)psoDst->pvScan0 
				+ (psoDst->sizlBitmap.cy - 1 - pRect->top) * (-screenstride)
				+ (pRect->left * pixelbits >> 3);
		}

		data = pCmd->cmdData;

		header = (MirrCmdHeader*)data;
		header->type = T_FB_GLYPH;
		header->dstId = psurfDst->id;
		header->srcId = NULL_SURF_ID;
		
		DISPDBG((TEXT_DBG_LEV, " Glyph: data allocation succeed. data_size = %d\n", datasize));

		fbglyph = (thinc_fbGlyph*)((char*)data + THINC_HEADERSIZE);
				
		fbglyph->numrects = 0;
		fbglyph->color = ulForeColor;
		fbglyph->x = pRect->left;
		fbglyph->y = pRect->top;
		fbglyph->width = width;
		fbglyph->height = height;
		

		bitmap = (char*)data + THINC_HEADERSIZE + T_FB_GLYPH_SIZE;

		//
		// Read the glyph information from the surface.
		//
		// FIXME: we can deal with 24-bit pixels only
		//
		ulForeColor &= 0x00ffffff;

		//DISPDBG((0, " bGetClippedGlyphs : height=%d, width=%d\n", height, width));


		for (i = 0; i < height; i++)
		{

			pulLine = (ULONG*)pscan;

			for (j = 0; j < alignedWidth; j++)
			{
				cBitsByte = 0;
				if ((*pulLine++ & 0x00ffffff) == ulForeColor) cBitsByte |= 0x01; 
				if ((*pulLine++ & 0x00ffffff) == ulForeColor) cBitsByte |= 0x02; 
				if ((*pulLine++ & 0x00ffffff) == ulForeColor) cBitsByte |= 0x04; 
				if ((*pulLine++ & 0x00ffffff) == ulForeColor) cBitsByte |= 0x08; 
				if ((*pulLine++ & 0x00ffffff) == ulForeColor) cBitsByte |= 0x10; 
				if ((*pulLine++ & 0x00ffffff) == ulForeColor) cBitsByte |= 0x20; 
				if ((*pulLine++ & 0x00ffffff) == ulForeColor) cBitsByte |= 0x40; 
				if ((*pulLine++ & 0x00ffffff) == ulForeColor) cBitsByte |= 0x80; 
				(*bitmap++) = cBitsByte;
			}
			
			if (remainingBits != 0)
			{
				bit = 1;
				cBitsByte = 0;
				for (j = 0; j < remainingBits; j++)
				{
					if (*pulLine++ == ulForeColor) 
					{
						cBitsByte |= bit; 
					}
					bit <<= 1;
				}
				//cBitsByte <<= remainingShift;
				(*bitmap++) = cBitsByte;
			}
			
			pscan = (PCHAR)pscan + screenstride;
			
		}

		bitmap = (char*)data + THINC_HEADERSIZE + T_FB_GLYPH_SIZE;

#ifdef TIME_STAMP
		// Time stamp the command
		ppdev->pfQuerySystemTime(&pCmd->timeStamp[0]);
#endif
		if (!thincPushCmdWithOffset(ppdev, ppdev->pCmdBlock, offset))
		{
			DISPDBG((1, "DrvTextOut: pushing command failed\n")); 
		}					   
	} 
	//DISPDBG((0, " bGetClippedGlyphs <-- \n"));
	return(TRUE);
}



//
// Push a CREATE_OFFSCREEN_BITMAP intermediate command into the command channel
//


BOOL CreateDeviceBitmap(PPDEV ppdev, PMirrSurf psurf)
{
	int					datasize, offset;
	PThincCmd			pCmd;
	MirrCmdHeader*		pHeader;

	datasize = THINC_HEADERSIZE + T_CREATE_OFFSURF_SIZE;

	pCmd = thincCmdBlockAlloc(&(ppdev->pCmdBlock), datasize, &offset);
	if (pCmd != NULL)
	{
		MirrCmdHeader*		header = (MirrCmdHeader*) pCmd->cmdData;
		thinc_createOffscreenSurf* pCreateBitmap = 
			(thinc_createOffscreenSurf*)((char*)(pCmd->cmdData) + THINC_HEADERSIZE);

		header->dstId = psurf->id;
		header->srcId = NULL_SURF_ID;
		header->type = T_CREATE_OFFSURF;
		
		pCreateBitmap->width = psurf->cx;
		pCreateBitmap->height = psurf->cy;
		pCreateBitmap->stride = psurf->lDelta;
		pCreateBitmap->depth = psurf->ulBitCount;
		pCreateBitmap->bpp = psurf->ulBitCount/8;
		pCreateBitmap->id = psurf->id;

#ifdef TIME_STAMP
		// Time stamp the command
		ppdev->pfQuerySystemTime(&pCmd->timeStamp[0]);
#endif
		return thincPushCmdWithOffset(ppdev, ppdev->pCmdBlock, offset);
	}
	else
	{
		return FALSE;
	}
}

//
// Push a DESTROY_OFFSCREEN_BITMAP intermediate command into the command channel
//

BOOL DestroyDeviceBitmap(PPDEV ppdev, int id)
{
	int					datasize, offset;
	PThincCmd			pCmd;

	datasize = THINC_HEADERSIZE + T_DESTROY_OFFSURF_SIZE;

	pCmd = thincCmdBlockAlloc(&(ppdev->pCmdBlock), datasize, &offset);
	if (pCmd != NULL)
	{
		MirrCmdHeader*		 header = (MirrCmdHeader*) pCmd->cmdData;
		thinc_destroyOffscreenSurf* pDestroyBitmap;

		header->dstId = id;
		header->srcId = NULL_SURF_ID;
		header->type = T_DESTROY_OFFSURF;

		pDestroyBitmap = (thinc_destroyOffscreenSurf*)((char*)(pCmd->cmdData) + THINC_HEADERSIZE);
		pDestroyBitmap->id = id;

#ifdef TIME_STAMP
		// Time stamp the command
		ppdev->pfQuerySystemTime(&pCmd->timeStamp[0]);
#endif
		return thincPushCmdWithOffset(ppdev, ppdev->pCmdBlock, offset);
	}
	return FALSE;
}

//
// Push a TIME_STAMP intermediate command into the command channel.
// The TIME_STAMP intermediat command is for instrument and debugging purpose.
//


BOOL TimeStamp(PPDEV ppdev, int type, char* str)
{
	int					datasize, offset;
	PThincCmd			pCmd;

	datasize = THINC_HEADERSIZE + T_TIME_STAMP_SIZE;

	pCmd = thincCmdBlockAlloc(&(ppdev->pCmdBlock), datasize, &offset);
	if (pCmd != NULL)
	{
		MirrCmdHeader*		 header = (MirrCmdHeader*) pCmd->cmdData;
		thinc_timeStamp*	 pTimeStamp;

		header->dstId = NULL_SURF_ID;
		header->srcId = NULL_SURF_ID;
		header->type = T_TIME_STAMP;

		pTimeStamp = (thinc_timeStamp*)((char*)(pCmd->cmdData) + THINC_HEADERSIZE);
		pTimeStamp->stampType = type;
		ppdev->pfQuerySystemTime(&(pTimeStamp->time));
		if (str != NULL)
		{
			strcpy(pTimeStamp->record,str);
		}
#ifdef TIME_STAMP
		// Time stamp the command
		ppdev->pfQuerySystemTime(&pCmd->timeStamp[0]);
#endif
		return thincPushCmdWithOffset(ppdev, ppdev->pCmdBlock, offset);
	}
	return FALSE;
}
