/* Copyright 2004, 2005 Lei Zhang, Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/******************************Module*Header*******************************\
*
*                           *********************
*                           * THINC driver code *
*                           *********************
*
* Module Name: mem.c
*
* contains the implementation of the loadable kernel dll. It includes the routines
* for event sharing, memory sharing between kernel mode and user mode. 
*
***************************************************************************/

#include <wdm.h>
//#include "comdef.h"

#define MEM_DBG_LEV 0

typedef struct _SharedMemoryBlockInfo {
	PVOID	kernelVirtualAddress;
	PVOID	userVirtualAddress;
	PMDL	mdl;
} SharedMemoryBlockInfo, *PSharedMemoryBlockInfo;

NTSTATUS
DriverEntry(
  IN PDRIVER_OBJECT DriverObject,
  IN PUNICODE_STRING RegistryPath
);

#ifdef ALLOC_PRAGMA
#pragma alloc_text(INIT, DriverEntry)
#endif

NTSTATUS
DllInitialize( IN PUNICODE_STRING pus )
{
    DbgPrint(MEM_DBG_LEV, "SAMPLE: DllInitialize(%S)\n", pus->Buffer);
    return STATUS_SUCCESS;
}

NTSTATUS
DllUnload( )
{
    DbgPrint((MEM_DBG_LEV, "SAMPLE: DllUnload\n"));
    return STATUS_SUCCESS;
}

// The DLL must have an entry point, but it is never called.
//
NTSTATUS
DriverEntry(
  IN PDRIVER_OBJECT DriverObject,
  IN PUNICODE_STRING RegistryPath
)
{
    return STATUS_SUCCESS;
}

//
// Create a block a memory that can be shared with user mode process
//
PVOID
CreateMem(IN OUT PVOID* pUserVirtualAddress,
		  IN OUT PVOID* pKernelVirtualAddress,
		  IN OUT PVOID* ppMemoryBlockInfo,
		  IN     SIZE_T	buffersize)

{
	PSharedMemoryBlockInfo pMemBlockInfo;
    
	PMDL	mdl;
	SIZE_T	memsize;
	memsize = buffersize / PAGE_SIZE + 1;
	
	DbgPrint((MEM_DBG_LEV, " mem!CreateMem: -->\n"));
	pMemBlockInfo = ExAllocatePoolWithTag(NonPagedPool, sizeof(SharedMemoryBlockInfo), 'MpaM');

	if (pMemBlockInfo == NULL)
	{
		return NULL;
	}
	
	*pKernelVirtualAddress = ExAllocatePoolWithTag(NonPagedPool, PAGE_SIZE * memsize, 'MpaM');

	if(!*pKernelVirtualAddress)
	{
		DbgPrint((MEM_DBG_LEV, " mem!CreateMemory: Fail to allocated nonpaged pool block\n"));
		return(NULL);
	}

	mdl = IoAllocateMdl(*pKernelVirtualAddress, PAGE_SIZE * memsize, FALSE, FALSE, NULL);

	if(!mdl) 
	{
		DbgPrint((MEM_DBG_LEV, " mem!CreateMemory: Fail to allocated MDL\n"));
		ExFreePool(*pKernelVirtualAddress);
		return(NULL);
	}

	MmBuildMdlForNonPagedPool(mdl);

	*pUserVirtualAddress = MmMapLockedPagesSpecifyCache(
							mdl, 
							UserMode,
							MmCached,
							NULL,
							FALSE,
							NormalPagePriority);

	if(!*pUserVirtualAddress) 
	{
		DbgPrint((MEM_DBG_LEV, " mem!CreateMemory: Fail to map the memory to user address\n"));
		IoFreeMdl(mdl);
		ExFreePool(*pKernelVirtualAddress);
		return(NULL);
	}

	pMemBlockInfo->kernelVirtualAddress = *pKernelVirtualAddress;
	pMemBlockInfo->userVirtualAddress = *pUserVirtualAddress;
	pMemBlockInfo->mdl=mdl;

	*ppMemoryBlockInfo = pMemBlockInfo;

	DbgPrint((MEM_DBG_LEV, " mem!CreateMem: <--\n"));

	return(pMemBlockInfo);

}

PVOID
DeAllocMem(IN OUT PVOID pMemoryBlockInfo)
{
	PSharedMemoryBlockInfo pMemBlockInfo = (PSharedMemoryBlockInfo)pMemoryBlockInfo;
	MmUnmapLockedPages(pMemBlockInfo->userVirtualAddress, pMemBlockInfo->mdl);
	IoFreeMdl(pMemBlockInfo->mdl);
	ExFreePool(pMemBlockInfo->kernelVirtualAddress);
	ExFreePool(pMemoryBlockInfo);
	return (NULL);
}

PVOID
MapMem(IN OUT PVOID* pUserVirtualAddress,
	   IN     PVOID  kernelVirtualAddress,
	   IN OUT PVOID* ppMemoryBlockInfo,
	   IN     SIZE_T	buffersize)

{
	PSharedMemoryBlockInfo pMemBlockInfo;
    
	PMDL	mdl;
	SIZE_T	memsize;
	memsize = buffersize / PAGE_SIZE + 1;
	
	//DbgPrint((MEM_DBG_LEV, " mem!MapMem: -->\n"));

	pMemBlockInfo = ExAllocatePoolWithTag(NonPagedPool, sizeof(SharedMemoryBlockInfo), 'MpaM');

	if (pMemBlockInfo == NULL)
	{
		return NULL;
	}
	
	ASSERT ((PAGE_SIZE * memsize < PAGE_SIZE * (65535 - sizeof(MDL)) / sizeof(ULONG_PTR), " mem!MapMem: memsize is over size "));

	mdl = IoAllocateMdl(kernelVirtualAddress, PAGE_SIZE * memsize, FALSE, FALSE, NULL);

	if(!mdl) 
	{
		DbgPrint((MEM_DBG_LEV, " mem!CreateMemory: Fail to allocated MDL\n"));
		return(NULL);
	}

	MmBuildMdlForNonPagedPool(mdl);

	*pUserVirtualAddress = MmMapLockedPagesSpecifyCache(
							mdl, 
							UserMode,
							MmCached,
							NULL,
							FALSE,
							NormalPagePriority);

	if(!*pUserVirtualAddress) 
	{
		DbgPrint((MEM_DBG_LEV, " mem!CreateMemory: Fail to map the memory to user address\n"));
		IoFreeMdl(mdl);
		return(NULL);
	}

	pMemBlockInfo->kernelVirtualAddress = kernelVirtualAddress;
	pMemBlockInfo->userVirtualAddress = *pUserVirtualAddress;
	pMemBlockInfo->mdl=mdl;

	*ppMemoryBlockInfo = pMemBlockInfo;

	//DbgPrint((MEM_DBG_LEV, " mem!MapMem: <--\n"));

	return(pMemBlockInfo);

}

PVOID
UnMapMem(IN OUT PVOID pMemoryBlockInfo)
{
	PSharedMemoryBlockInfo pMemBlockInfo = (PSharedMemoryBlockInfo)pMemoryBlockInfo;
	PVOID ret = pMemBlockInfo->kernelVirtualAddress;
	//DbgPrint((MEM_DBG_LEV, " mem!UnMapMem: -->\n"));
	MmUnmapLockedPages(pMemBlockInfo->userVirtualAddress, pMemBlockInfo->mdl);
	IoFreeMdl(pMemBlockInfo->mdl);
	ExFreePool(pMemoryBlockInfo);
	//DbgPrint((MEM_DBG_LEV, " mem!UnMapMem: <--\n"));
	return (ret);
}

PVOID
OpenAndSetSharedEvent(IN PCWSTR eventName)
{
	PKEVENT			pSharedEvent = NULL;
	HANDLE			hSharedEvent = NULL;
	UNICODE_STRING	uEventName;
	NTSTATUS		res; 

	DbgPrint((0, " mem!OpenAndSetSharedEvent: -->\n"));

	RtlInitUnicodeString(&uEventName, eventName);

	//
	// Get a pointer to a KEVENT object that is associated with this
	// named event
	//
	pSharedEvent = IoCreateNotificationEvent(&uEventName, &hSharedEvent);
		
	if (pSharedEvent != NULL) 
	{ 
		DbgPrint((0, " mem!OpenAndSetSharedEvent: EventOpened\n"));

		//
		// Take a reference out on the object so that 
		//  it does not go away if the application
		//  goes away unexpectedly
		//
		ObReferenceObject(pSharedEvent);
		//
		// Set the event to nonsignaled state
		//
		KeSetEvent(pSharedEvent, 0, FALSE);
	}  

	DbgPrint((0, " mem!OpenAndSetSharedEvent: <--\n"));
	
	return pSharedEvent;
}

VOID
CloseSharedEvent(IN PVOID pEvent)
{
	DbgPrint((0, " mem!CloseSharedEvent: -->\n"));
	ObDereferenceObject(pEvent);
	DbgPrint((0, " mem!CloseSharedEvent: <--\n"));
}

VOID 
WaitOnSharedEvent(IN PVOID pEvent,
				  IN PLARGE_INTEGER  Timeout)
{
	//DbgPrint((0, " mem!WaitOnSharedEvent: -->\n"));
	
	KeWaitForSingleObject(pEvent,
						  UserRequest,
						  KernelMode,
						  TRUE,
						  Timeout);

	//DbgPrint((0, " mem!WaitOnSharedEvent: <--\n"));
}

LONG
SetSharedEvent(IN PVOID pEvent)
{
	return KeSetEvent(pEvent, 0, FALSE);
}

LONG
ResetSharedEvent(IN PVOID pEvent)
{
	return KeResetEvent(pEvent);
}

LONG
AtomicCompareExchange(
    IN OUT PLONG  Destination,
    IN LONG  Exchange,
    IN LONG  Comparand
    )
{
	return InterlockedCompareExchange(Destination, Exchange, Comparand);
}

VOID QuerySystemTime(
	 OUT PVOID pTime
	 )
{
	LARGE_INTEGER time;
	KeQuerySystemTime(&time);
	RtlTimeToTimeFields(&time, (PTIME_FIELDS) pTime);
}
